// Author(s): Thomas Neele
// Copyright: see the accompanying file COPYING or copy at
// https://github.com/mCRL2org/mCRL2/blob/master/COPYING
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)
//
/// \file partition.h


#ifndef MCRL2_PBESSYMBOLICBISIM_PARTITION_H
#define MCRL2_PBESSYMBOLICBISIM_PARTITION_H

#include <string>
#include <queue>
#include <unordered_map>
#include <unordered_set>

#include "mcrl2/bes/boolean_equation_system.h"
#include "mcrl2/bes/join.h"
#include "mcrl2/data/bool.h"
#include "mcrl2/data/find.h"
#include "mcrl2/data/fourier_motzkin.h"
#include "mcrl2/data/lambda.h"
#include "mcrl2/data/replace.h"
#include "mcrl2/data/rewriter.h"
#include "mcrl2/data/rewriters/one_point_rule_rewriter.h"
#include "mcrl2/data/rewriters/quantifiers_inside_rewriter.h"
#include "mcrl2/pbes/detail/ppg_rewriter.h"
#include "mcrl2/pg/ParityGame.h"
#include "mcrl2/smt/cvc4.h"
#include "mcrl2/smt/solver.h"
#include "mcrl2/smt/translation_error.h"
#include "mcrl2/utilities/logger.h"

#include "simplifier.h"
#define THIN       "0"
#define BOLD       "1"
#define GREEN(S)  "\033[" S ";32m"
#define YELLOW(S) "\033[" S ";33m"
#define RED(S)    "\033[" S ";31m"
#define NORMAL    "\033[0;0m"
#include "ppg_parser.h"
#include "block.h"


namespace mcrl2
{
namespace data
{

class dependency_graph_partition
{
  typedef block block_t;
  typedef pbes_system::detail::ppg_equation equation_type_t;

protected:
  const pbes_system::detail::ppg_pbes m_spec;

  data_manipulators m_dm;
  bool m_early_termination;

  std::list< block_t >      m_proof_blocks;
  std::list< block_t >      m_other_blocks;
  std::map< block_t, block_t > m_strategy;

  split_cache<block>* m_block_cache;
  split_cache<subblock>* m_subblock_cache;
  std::map< pbes_system::detail::ppg_equation, priority_t > m_rank_map;

  bool is_valid_approximation(const block_t& src, bool is_positive_pg) const
  {
    return (src.is_conjunctive() && is_positive_pg) || (!src.is_conjunctive() && !is_positive_pg);
  }

  bool is_in_strategy(const block_t& src, const block_t& dest) const
  {
    auto find_result = m_strategy.find(src);
    // If there is no winning strategy in src, then m_strategy does not contain
    // src and all outgoing transitions from src have to be considered.
    return find_result == m_strategy.end() || find_result->second == dest;
  }

  /**
   * \brief Split block phi_k on phi_l
   */
  bool split_block(const block_t& phi_k, const block_t& phi_l, const std::vector<subblock>& subblocks, bool use_optimisations)
  {
    if(use_optimisations && !is_in_strategy(phi_k, phi_l))
    {
      // This edge is not part of the winning strategy in the parity game
      return false;
    }
    if(m_block_cache->check_refinement(phi_k, phi_l))
    {
      // This split has already been tried before
      return false;
    }
    if(!phi_k.has_transition(phi_l))
    {
      // There is no transition between phi_k and phi_l, so a split will definitely
      // not succeed
      return false;
    }

    mCRL2log(log::verbose) << "Splitting\n" << phi_k << "wrt\n" << phi_l << std::endl;

    // Try to split
    const std::pair<block, block> split_result = phi_k.split(phi_l, subblocks, use_optimisations);
    if(split_result.first.is_empty() || split_result.second.is_empty())
    {
      // Split failed
      return false;
    }

    if(!use_optimisations)
    {
      // Update the caches and the partition
      block_t phi_k_copy(phi_k);
      block_t phi_l_copy(phi_l);
      m_proof_blocks.remove(phi_k);
      std::vector<std::list<block>> block_vec({m_proof_blocks, m_other_blocks});
      elements_iterator< block, std::vector<std::list<block>>> block_it(block_vec);
      m_block_cache->replace_after_split(block_it, block_it.end(), phi_k_copy, split_result.first, split_result.second);
      m_block_cache->insert_transition(split_result.first, phi_l_copy, true);
      m_block_cache->insert_transition(split_result.second, phi_l_copy, false);

      m_proof_blocks.push_front(split_result.first);
      m_proof_blocks.push_front(split_result.second);
    }
    return true;
  }

  std::vector<subblock> make_subblock_list() const
  {
    std::vector<subblock> result;
    for(const block_t& b: m_proof_blocks)
    {
      result.insert(result.end(), b.subblocks().begin(), b.subblocks().end());
    }
    for(const block_t& b: m_other_blocks)
    {
      result.insert(result.end(), b.subblocks().begin(), b.subblocks().end());
    }
    return result;
  }

  /**
   * \brief Try every combination of blocks to
   * see whether the partition can be refined.
   */
  bool refine_step(bool use_optimisations, bool is_positive_pg)
  {
    std::vector<subblock> subblocks = make_subblock_list();
    // Sorting the blocks within the BFS layers such that the following splits will be tried first:
    // - In the case of a negative proof graph, conjunctive nodes are split first
    // - In the case of a positive proof graph, disjunctive nodes are split first
    m_proof_blocks.sort([is_positive_pg](const block_t& b1, const block_t& b2)
                          {
                            return b1.bfs_level < b2.bfs_level || (b1.bfs_level == b2.bfs_level && ((is_positive_pg && !b1.is_conjunctive() && b2.is_conjunctive()) || (!is_positive_pg && b1.is_conjunctive() && !b2.is_conjunctive())));
                          });
    for(const block_t& phi_k: m_proof_blocks)
    {
      if(use_optimisations && is_valid_approximation(phi_k, is_positive_pg))
      {
        continue;
      }
      // Sort the blocks within their BFS layers such that edges present in the proof graph
      // are tried as splitters first.
      std::list<block_t> proof_blocks(m_proof_blocks);
      proof_blocks.sort([&](const block_t& b1, const block_t& b2)
                          {
                            return b1.bfs_level < b2.bfs_level || (b1.bfs_level == b2.bfs_level && is_in_strategy(phi_k,b1) && !is_in_strategy(phi_k,b2));
                          });
      for(const block_t& phi_l: m_proof_blocks)
      {
        if(split_block(phi_k, phi_l, subblocks, use_optimisations))
        {
          return true;
        }
      }
    } // stop when we have investigated all blocks
    return false;
  }

  /**
   * \brief Finds the block in blocks that contains the initial state
   */
  template <typename Container>
  block_t find_initial_block(const Container& blocks, typename atermpp::enable_if_container<Container, block_t>::type* = nullptr) const
  {
    for(const block_t& block: blocks)
    {
      if(block.contains_state(m_spec.initial_state()))
      {
        return block;
      }
    }
    throw mcrl2::runtime_error("Initial block not found.");
  }

  /**
   * \brief Computes the reachable blocks.
   * \detail The partition is updated, such that it only contains
   * reachable blocks.
   * \param only_check_proof_blocks Set to true when reachability should only be checked
   * within m_proof_blocks
   */
  void find_reachable_blocks(const bool& only_check_proof_blocks)
  {
    std::list<block_t> unreachable(m_proof_blocks.begin(),m_proof_blocks.end());
    if(!only_check_proof_blocks)
    {
      unreachable.insert(unreachable.end(), m_other_blocks.begin(), m_other_blocks.end());
    }
    std::queue<block_t> open_set;

    // Search for the block that contains the initial state
    block_t initial_block = find_initial_block(unreachable);
    initial_block.bfs_level = 0;
    unreachable.erase(std::find(unreachable.begin(), unreachable.end(), initial_block));
    open_set.push(initial_block);
    m_proof_blocks.clear();
    if(!only_check_proof_blocks)
    {
      m_other_blocks.clear();
    }
    m_proof_blocks.push_back(initial_block);

    while(!open_set.empty())
    {
      // Take a block from the open set
      const block_t block = open_set.front();
      open_set.pop();

      // Look for possible successors of block in the set of unreachable blocks
      for(std::list<block_t>::iterator i = unreachable.begin(); i != unreachable.end();)
      {
        if(block.has_transition(*i))
        {
          i->bfs_level = block.bfs_level + 1;
          // A transition was found, so *i is reachable
          open_set.push(*i);
          m_proof_blocks.push_back(*i);
          i = unreachable.erase(i);
        }
        else
        {
          ++i;
        }
      }
    }


    mCRL2log(log::verbose) << RED(THIN) << "Unreachable blocks:" << NORMAL << std::endl;
    int i = 0;
    for(const block_t& block: unreachable)
    {
      if(only_check_proof_blocks)
      {
        // Add the unreachable blocks to m_other_blocks to make
        // sure no blocks get lost
        // When we only checking reachability in m_proof_blocks,
        // the unreachable blocks might still be reachable through
        // some blocks in m_other_blocks
        // Only when we are checking reachability in the whole graph,
        // we can really throw away the unreachable blocks.
        m_other_blocks.push_front(block);
      }
      // split_logger->mark_deleted(rewr(block));
      mCRL2log(log::verbose) << "  block " << i << "\n" << pp(block);
      i++;
    }
  }

  /**
   * \brief Build the BES that follows from the current
   * partition.
   * \detail Build the BES that contains one variable for each
   * block in the partition. If the partition is stable,
   * then this BES is bisimilar to the original PBES.
   */
  // template <typename Container>
  // bes::boolean_equation_system make_bes(const Container& blocks, typename atermpp::enable_if_container<Container, block_t>::type* = nullptr)
  // {
  //   //TODO fix bug where variables are sorted according to BFS layer and not according
  //   // to the order in the original PBES.
  //   // This may result in a BES with a different solution.
  //   set_identifier_generator id_gen;
  //   std::map< block_t, bes::boolean_variable > var_map;
  //   for(const block_t& block: blocks)
  //   {
  //     var_map.insert(std::make_pair(block, bes::boolean_variable(id_gen(std::string(block.name())))));
  //   }
  //
  //   std::vector< bes::boolean_equation > equations;
  //   for(const block_t& src: blocks)
  //   {
  //     std::set<bes::boolean_expression> right_hand_side;
  //     for(const block_t& dest: blocks)
  //     {
  //       if(src.has_transition(dest))
  //       {
  //         right_hand_side.insert(var_map[dest]);
  //       }
  //     }
  //
  //     bes::boolean_expression rhs_expr = src.is_conjunctive() ? bes::join_and(right_hand_side.begin(), right_hand_side.end()) : bes::join_or(right_hand_side.begin(), right_hand_side.end());
  //     equations.emplace_back(src.fixpoint_symbol(), var_map[src], rhs_expr);
  //   }
  //
  //   // Search for the initial block
  //   const block_t& initial_block = find_initial_block(blocks);
  //   return bes::boolean_equation_system(equations, var_map[initial_block]);
  // }

  void make_rank_map()
  {
    pbes_system::fixpoint_symbol previous_symbol(pbes_system::fixpoint_symbol::nu());
    priority_t rank = 0;
    for(const equation_type_t& eq: m_spec.equations())
    {
      if(eq.symbol() != previous_symbol)
      {
        previous_symbol = eq.symbol();
        rank++;
      }
      m_rank_map.insert(std::make_pair(eq, rank));
    }
  }

  inline
  void make_pg(ParityGame& pg, const std::list<block>& blocks) const
  {
    StaticGraph::edge_list edges;
    ParityGameVertex* node_attributes = new ParityGameVertex[blocks.size()];
    int src_index = 0;
    for(const auto& src: blocks)
    {
      node_attributes[src_index].player = src.is_conjunctive() ? player_t::PLAYER_ODD : player_t::PLAYER_EVEN;
      node_attributes[src_index].priority = src.rank(m_rank_map);

      bool has_outgoing_edge = false;
      int dest_index = 0;
      for(const auto& dest: blocks)
      {
        if(src.has_transition(dest))
        {
          edges.emplace_back(src_index, dest_index);
          has_outgoing_edge = true;
        }
        dest_index++;
      }

      if(!has_outgoing_edge)
      {
        // All nodes should have an outgoing edge, since SRF requires
        // a summand to either X_true or X_false in every right-hand side
        throw mcrl2::runtime_error("Did not find an outgoing edge for the block " + pp(src));
      }
      src_index++;
    }

    StaticGraph underlying_graph;
    underlying_graph.assign(edges, StaticGraph::EdgeDirection::EDGE_BIDIRECTIONAL);
    pg.assign(underlying_graph, node_attributes);
  }

  /**
   * \brief Print the current partition to stdout
   */
  template <typename Container>
  void print_partition(const Container& blocks, typename atermpp::enable_if_container<Container, block_t>::type* = nullptr)
  {
    int i = 0;
    for(const block_t& block: blocks)
    {
      mCRL2log(log::verbose) << YELLOW(THIN) << "  block " << i << "\n" << NORMAL << block;
      ++i;
    }
  }

  void make_initial_partition(bool fine_initial)
  {
    if(fine_initial)
    {
      // Create on block for every equation
      for(const equation_type_t& eq: m_spec.equations())
      {
        m_proof_blocks.emplace_back(std::list<subblock>({subblock(eq, m_dm, m_subblock_cache)}), m_block_cache);
      }
    }
    else
    {
      // For each PBES rank, create one block with conjunctive subblocks and one
      // block with disjunctive subblocks.
      pbes_system::fixpoint_symbol previous_symbol;
      std::list<subblock> conjunctive_subblocks;
      std::list<subblock> disjunctive_subblocks;
      for(std::vector<equation_type_t>::const_iterator it = m_spec.equations().begin(); it != m_spec.equations().end();)
      {
        previous_symbol = it->symbol();
        (it->is_conjunctive() ? conjunctive_subblocks : disjunctive_subblocks).emplace_back(*it, m_dm, m_subblock_cache);

        ++it;
        if(it == m_spec.equations().end() || it->symbol() != previous_symbol)
        {
          if(!conjunctive_subblocks.empty())
          {
            m_proof_blocks.emplace_back(conjunctive_subblocks, m_block_cache);
            conjunctive_subblocks = std::list<subblock>();
          }
          if(!disjunctive_subblocks.empty())
          {
            m_proof_blocks.emplace_back(disjunctive_subblocks, m_block_cache);
            disjunctive_subblocks = std::list<subblock>();
          }
        }
      }
    }
  }

public:

  dependency_graph_partition(const pbes_system::detail::ppg_pbes& spec, const rewriter& r, const rewriter& pr, const simplifier_mode& simpl_mode, bool fine_initial, bool early_termination)
  : m_spec(spec)
  , m_dm(r, pr, spec.data())
  , m_early_termination(early_termination)
  {
    m_dm.contains_reals = false;
    for(const equation_type_t& eq: m_spec.equations())
    {
      m_dm.simpl->insert(std::make_pair(eq.variable(), get_simplifier_instance(simpl_mode, m_dm.rewr, m_dm.proving_rewr, eq.variable().parameters(), m_spec.data())));
      m_dm.contains_reals |= std::find_if(eq.variable().parameters().begin(), eq.variable().parameters().end(),
        [](const variable& var) { return var.sort() == sort_real::real_();}) != eq.variable().parameters().end();
    }
    m_block_cache = new split_cache<block>();
    m_subblock_cache = new split_cache<subblock>();
    make_initial_partition(fine_initial);
    make_rank_map();
    print_partition(m_proof_blocks);
    // Make sure the initial block is the first in the list
    const block_t& initial_block = find_initial_block(m_proof_blocks);
    m_proof_blocks.erase(std::find(m_proof_blocks.begin(), m_proof_blocks.end(), initial_block));
    m_proof_blocks.push_front(initial_block);
  }

  std::list< block_t > get_proof_blocks()
  {
    return m_proof_blocks;
  }

  std::list< block_t > get_other_blocks()
  {
    return m_other_blocks;
  }

  void set_proof(const std::set< std::size_t >& proof_nodes, const std::vector< std::size_t >& strategy)
  {
    // Move all the blocks to the other set while keeping the order m_proof_block ++ m_other_blocks
    std::vector< block_t > all_blocks(m_proof_blocks.begin(), m_proof_blocks.end());
    // all_blocks.swap(m_proof_blocks);
    all_blocks.insert(all_blocks.end(), m_other_blocks.begin(), m_other_blocks.end());
    m_proof_blocks.clear();
    m_other_blocks.clear();

    // Move the requested blocks to the main set
    std::size_t i = 0;
    m_strategy.clear();
    for(const block_t& b: all_blocks)
    {
      if(proof_nodes.find(i) != proof_nodes.end())
      {
        m_proof_blocks.push_back(b);
        if(strategy[i] != (std::size_t) -1)
        {
          // There is a winning strategy in node i by going to strategy[i]
          m_strategy.insert(std::make_pair(b, all_blocks[strategy[i]]));
        }
      }
      else
      {
        m_other_blocks.push_back(b);
      }
      i++;
    }
  }

  void get_reachable_pg(ParityGame& pg)
  {
    make_pg(pg, m_proof_blocks);
  }

  void print()
  {
    mCRL2log(log::verbose) << GREEN(THIN) << "Partition proof blocks:" << NORMAL << std::endl;
    print_partition(m_proof_blocks);
    mCRL2log(log::verbose) << GREEN(THIN) << "Partition other blocks:" << NORMAL << std::endl;
    print_partition(m_other_blocks);
  }

  // void save_bes()
  // {
  //   const bes::boolean_equation_system& bes = make_bes(m_proof_blocks);
  //   std::ofstream out;
  //   out.open("out.bes");
  //   bes.save(out);
  //   out.close();
  // }

  /**
   * Refine the partition num_steps times
   * When num_steps is 0, the algorithm only
   * terminates when the partition is stable.
   * \return true when the partition was already stable
   */
  bool refine_n_steps(std::size_t num_steps, bool is_positive_pg)
  {
    // mCRL2log(log::verbose) << "Initial partition:" << std::endl;
    // print_partition(m_proof_blocks);

    // First check whether the current proof graph is stable under optimisations.
    // This will not affect the cache.
    if(m_early_termination && !refine_step(true, is_positive_pg))
    {
      return true;
    }
    std::size_t num_iterations = 0;
    while(refine_step(false, is_positive_pg))
    {
      print();
      num_iterations++;
      if(num_steps != 0 && num_iterations >= num_steps)
      {
        break;
      }
      // if(num_steps == 0)
      // {
      //   make_bes(m_proof_blocks);
      // }
      // Check reachability only in m_proof_blocks.
      // Blocks that are not reachable in there will
      // be moved to m_other_blocks, because it is not
      // certain that they are unreachable when considering
      // the whole graph.
      find_reachable_blocks(true);
      mCRL2log(log::verbose) << "End of a refinement step " << num_iterations << std::endl;
    }
    // Check reachability in the full graph.
    // In this case, unreachable blocks will really be
    // thrown away.
    find_reachable_blocks(false);
    if(num_iterations == 0 || num_steps == 0)
    {
      mCRL2log(log::verbose) << "Final partition:" << std::endl;
      print_partition(m_proof_blocks);
    }
    return num_iterations == 0;
  }

};

} // namespace data
} // namespace mcrl2


#endif // MCRL2_PBESSYMBOLICBISIM_PARTITION_H
