// Author(s): Thomas Neele
// Copyright: see the accompanying file COPYING or copy at
// https://svn.win.tue.nl/trac/MCRL2/browser/trunk/COPYING
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)
//
/// \file simplifier_dbm.h


#ifndef MCRL2_LPSSYMBOLICBISIM_SIMPLIFIER_DBM_H
#define MCRL2_LPSSYMBOLICBISIM_SIMPLIFIER_DBM_H

#include "mcrl2/data/detail/linear_inequalities_utilities.h"

#include "base/bitstring.h"
#include "dbm/dbm.h"
#include "dbm/fed.h"

#include "simplifier.h"
#include "simplifier_mdd.h"

namespace mcrl2
{
namespace data
{

class simplifier_dbm: public simplifier
{
  typedef simplifier super;
  using super::rewr;
  using super::proving_rewr;

protected:
  const std::vector< variable > real_process_parameters;
  const uint32_t dim;
  data_specification dataspec;
  simplifier_mdd simpl_discr;

  const bool use_extrapolation;
  int32_t* lower_bounds;
  int32_t* upper_bounds;

  template< class InputIt >
  int get_dbm_index(const variable& var, InputIt first, const InputIt& last)
  {
    // First element in DBM is 0
    for(int i = 1; first != last; first++, i++)
    {
      if(var == *first)
      {
        return i;
      }
    }
    throw mcrl2::runtime_error("Variable " + pp(var) + " not found.");
    return -1;
  }

  std::pair< cindex_t, cindex_t > dbm_index_pair(const detail::lhs_t& pars, bool invert)
  {
    if(pars.size() > 2)
    {
      throw mcrl2::runtime_error("Expressions with more than 2 variables are not supported: " + detail::pp(pars));
    }
    cindex_t neg_index = 0;
    cindex_t pos_index = 0;
    for(const detail::variable_with_a_rational_factor& p: pars)
    {
      if(p.factor() == real_one())
      {
        pos_index = get_dbm_index(p.variable_name(), real_process_parameters.begin(), real_process_parameters.end());
      }
      else if(p.factor() == real_minus_one())
      {
        neg_index = get_dbm_index(p.variable_name(), real_process_parameters.begin(), real_process_parameters.end());
      }
      else
      {
        throw mcrl2::runtime_error("Variable " + pp(p.variable_name()) + " has a factor " + pp(p.factor()));
      }
    }
    std::pair< cindex_t, cindex_t > result(pos_index, neg_index);
    if(invert) {
      std::swap(result.first, result.second);
    }
    return result;
  }

  int32_t int_from_expr(const data_expression& expr)
  {
    static std::map<data_expression, int32_t> cache;
    auto find_result = cache.find(expr);
    if(find_result != cache.end())
    {
      return find_result->second;
    }
    int result;
    if(expr == sort_pos::c1())
    {
      result = 1;
    }
    else if(expr == sort_nat::c0())
    {
      result = 0;
    }
    else if(sort_int::is_cneg_application(expr))
    {
      result = -1 * int_from_expr(sort_int::arg(expr));
    }
    else if(sort_int::is_cint_application(expr))
    {
      result = int_from_expr(sort_int::arg(expr));
    }
    else if(sort_nat::is_cnat_application(expr))
    {
      result = int_from_expr(sort_nat::arg(expr));
    }
    else if(sort_pos::is_cdub_application(expr))
    {
      result = 2 * int_from_expr(sort_pos::right(expr)) + (sort_pos::left(expr) == sort_bool::true_() ? 1 : 0);
    }
    else
    {
      throw mcrl2::runtime_error("Encountered unexpected expression while traversing integer number: " + pp(expr));
    }
    cache.insert(std::make_pair(expr, result));
    return result;
  }

  void add_linear_inequality(dbm::dbm_t& dbm, const linear_inequality& li)
  {
    std::pair< cindex_t, cindex_t > indices = dbm_index_pair(li.lhs(), false);
    dbm.constrain(indices.first, indices.second, int_from_expr(rewr(sort_real::real2int(li.rhs()))), li.comparison() == detail::less);
    if(li.comparison() == detail::equal)
    {
      dbm.constrain(indices.second, indices.first, -int_from_expr(rewr(sort_real::real2int(li.rhs()))), false);
    }
  }

  void add_zone(dbm::fed_t& fed, const data_expression_list cons)
  {
    dbm::dbm_t dbm(dim);
    dbm.setInit();
    for(const data_expression& expr: cons)
    {
      add_linear_inequality(dbm, linear_inequality(expr, rewr));
    }
    fed |= dbm;
  }

  data_expression add_zero_bounds(dbm::dbm_t& dbm, const data_expression& expr)
  {
    data_expression result = expr;
    const raw_t *raw_dbm = dbm.getDBM();
    for(cindex_t i = 1; i < dbm.getDimension(); i++)
    {
      if(dbm(0,i) >= dbm_LE_ZERO && dbm_addFiniteRaw(raw_dbm[i], raw_dbm[i*dim]) != dbm_LE_ZERO)
      {
        result = lazy::and_(result, greater_equal(real_process_parameters[i-1], real_zero()));
      }
    }
    return result;
  }

  data_expression dbm_to_expr(dbm::dbm_t dbm1)
  {
    if (dbm1.isEmpty())
    {
        return sort_bool::false_();
    }

    data_expression result = sort_bool::true_();
    uint32_t mingraph[bits2intsize(dim*dim)];
    const raw_t *dbm = dbm1.getDBM();
    size_t n = dbm_cleanBitMatrix(dbm, dim, mingraph,
                                  dbm_analyzeForMinDBM(dbm, dim, mingraph));

    for(cindex_t i = 0; i < dim; ++i)
    {
      for(cindex_t j = 0; j < dim; ++j)
      {
        if (base_readOneBit(mingraph, i*dim+j))
        {
          assert(i != j); // Not part of mingraph.

          // Write constraints xi-xj <|<= DBM[i,j]
          if (i == 0)
          {
            assert(j != 0); // 0,0 never part of mingraph
            if (dbm_addFiniteRaw(dbm[j], dbm[j*dim]) == dbm_LE_ZERO)
            {
              assert(n);
              n -= base_getOneBit(mingraph, j*dim);
              base_resetOneBit(mingraph, j*dim);

              // xj == b
              result = lazy::and_(result, equal_to(real_process_parameters[j-1], sort_real::real_(-dbm_raw2bound(dbm[j]))));
            }
            else // b < xj, b <= xj. Translated as -xj < -b and -xj <= -b
            {
              result = lazy::and_(result,
                application(dbm_rawIsStrict(dbm[j]) ? less(sort_real::real_()) : less_equal(sort_real::real_()),
                sort_real::negate(real_process_parameters[j-1]), sort_real::real_(dbm_raw2bound(dbm[j]))));
            }
          }
          else
          {
            function_symbol op;
            // xi-xj == b ?
            if (dbm_addFiniteRaw(dbm[i*dim+j], dbm[j*dim+i]) == dbm_LE_ZERO)
            {
              assert(n);
              n -= base_getOneBit(mingraph, j*dim+i);
              base_resetOneBit(mingraph, j*dim+i);

              // xi == xj
              if (j > 0 && dbm[i*dim+j] == dbm_LE_ZERO)
              {
                result = lazy::and_(result, equal_to(real_process_parameters[i-1], real_process_parameters[j-1]));
                goto consume_n;
              }
              // == b
              op = equal_to(sort_real::real_());
            }
            else
            {
              // xi < xj, xi <= xj
              if (j > 0 && dbm_raw2bound(dbm[i*dim+j]) == 0)
              {
                result = lazy::and_(result,
                  application(dbm_rawIsStrict(dbm[i*dim+j]) ? less(sort_real::real_()) : less_equal(sort_real::real_()),
                  real_process_parameters[i-1], real_process_parameters[j-1]));
                goto consume_n;
              }
              // < b, <= b
              op = dbm_rawIsStrict(dbm[i*dim+j]) ? less(sort_real::real_()) : less_equal(sort_real::real_());
            }
            data_expression rhs(sort_real::real_(dbm_raw2bound(dbm[i*dim+j])));

            if (j == 0) // xi < b, xi <= b
            {
              assert(i != 0); // 0,0 never part of mingraph
              result = lazy::and_(result, application(op, real_process_parameters[i-1], rhs));
            }
            else // xi-xj < b, xi-xj <= b
            {
              result = lazy::and_(result, application(op, sort_real::minus(real_process_parameters[i-1], real_process_parameters[j-1]), rhs));
            }
          }
        consume_n:
          assert(n);
          --n; // Consumed one constraint.
          if (!n) goto stop_loops;
        }
      }
    }
    stop_loops:
    return add_zero_bounds(dbm1, result);
    // return result;
  }

  data_expression fed_to_expr(const dbm::fed_t& fed)
  {
    data_expression result = sort_bool::false_();
    for (const dbm::dbm_t& dbm : fed)
    {
      result = lazy::or_(result, dbm_to_expr(dbm));
    }
    return result;
  }

  void extrapolate(dbm::fed_t& fed)
  {
    if(use_extrapolation)
    {
      fed.diagonalExtrapolateLUBounds(lower_bounds, upper_bounds);
    }
  }

  data_expression simplify_expression(const data_expression& expr)
  {
    // std::cout << "Simplifying expression " << expr << std::endl;
    std::vector < data_expression_list > real_conditions;
    std::vector < data_expression > non_real_conditions;
    detail::split_condition(expr, real_conditions, non_real_conditions);
    std::map< data_expression, std::vector< data_expression_list >> discr_to_real;
    for(uint32_t i = 0; i < real_conditions.size(); i++)
    {
      std::pair< std::map< data_expression, std::vector< data_expression_list >>::iterator, bool > res =
          discr_to_real.insert(std::make_pair(simpl_discr.apply(non_real_conditions[i]), std::vector< data_expression_list >()));
      res.first->second.push_back(real_conditions[i]);
    }
    data_expression result = sort_bool::false_();
    for(const std::pair< data_expression, std::vector< data_expression_list >>& disjunct: discr_to_real)
    {
      dbm::fed_t fed(dim);
      for(const data_expression_list& zone: disjunct.second)
      {
        add_zone(fed, zone);
      }
      fed = fed.partitionReduce();

      // optional step
      extrapolate(fed);

      result = lazy::or_(result, lazy::and_(disjunct.first, fed_to_expr(fed)));
    }
    return rewr(result);
  }

  /// \brief Returns a list of all real variables in l
  /// \param l a list of data variables
  /// \return The list of all v in l such that v.sort() == real()
  inline
  std::vector< variable > get_real_variables(const variable_list& l)
  {
    std::vector< variable > r;
    for (const variable& i: l)
    {
      if (i.sort() == sort_real::real_())
      {
        r.push_back(i);
      }
    }
    return r;
  }

  void make_lu_arrays(const std::map< variable, std::pair< data_expression, data_expression > >& lu_map)
  {
    if(!lu_map.empty())
    {
      lower_bounds = new int32_t[real_process_parameters.size() + 1];
      upper_bounds = new int32_t[real_process_parameters.size() + 1];
      lower_bounds[0] = upper_bounds[0] = 0;
      int i = 1;
      for(const variable& v: real_process_parameters)
      {
        const std::pair<data_expression, data_expression>& bounds = lu_map.at(v);
        lower_bounds[i] = bounds.first != data_expression() ? int_from_expr(rewr(sort_real::real2int(bounds.first))) : -dbm_INFINITY;
        upper_bounds[i] = bounds.second != data_expression() ? int_from_expr(rewr(sort_real::real2int(bounds.second))) : -dbm_INFINITY;
        i++;
      }
    }
  }

public:
  simplifier_dbm(const rewriter& r, const rewriter& pr, const variable_list& process_pars, const data_specification& dataspec_,
    const std::map< variable, std::pair<data_expression, data_expression> >& lu_map = std::map< variable, std::pair<data_expression, data_expression> >())
  : super(r, pr)
  , real_process_parameters(get_real_variables(process_pars))
  , dim(real_process_parameters.size() + 1)
  , dataspec(dataspec_)
  , simpl_discr(r, pr, dataspec_)
  , use_extrapolation(!lu_map.empty())
  {
    //optional step
    make_lu_arrays(lu_map);
  }

};


} // namespace mcrl2
} // namespace data

#endif // MCRL2_LPSSYMBOLICBISIM_SIMPLIFIER_DBM_H
